/*
 *
 * Copyright (C) 2009,2010,2011 Halon Security <support@halon.se>
 *
 * This file is part of libdkim++.
 *
 * libdkim++ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libdkim++ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with libdkim++.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "Resolver.hpp"

#if ! defined EMPTY_RESOLVER
#if defined _MSC_VER
#include <winsock2.h>
#include <WinDNS.h>
#else //defined _MSC_VER
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <memory.h>
#endif //defined _MSC_VER

#if defined _MACOS
#include <arpa/nameser_compat.h> 
#endif //defined _MACOS
#endif //! defined EMPTY_RESOLVER

namespace DKIM {
	namespace Util {

#if defined _MSC_VER
Resolver::Resolver()
{
}

Resolver::~Resolver()
{
}

/*
 * request for the T_TXT record of an domain name, if an error occures (false is returned)
 *  else true is returned (regardless if the domain txt record exists or not)
 */
bool Resolver::GetTXT(const std::string& domain, std::string& result)
{
#if ! defined EMPTY_RESOLVER
	PDNS_RECORD pdns = 0;

	long res = DnsQuery(domain.c_str(), DNS_TYPE_TEXT, NULL, NULL, &pdns, NULL);
	if (res != DNS_RCODE_NOERROR) {
		DnsRecordListFree(pdns, DnsFreeRecordList);
		return false;
	}

	PDNS_RECORD pdnsIt = pdns;
	while (pdnsIt != 0) {
		if (pdnsIt != 0 && pdnsIt->wType == DNS_TYPE_TEXT) {
			for (unsigned i = 0; i < pdnsIt->Data.TXT.dwStringCount; ++i) {
				if (pdnsIt->Data.TXT.pStringArray[i] != 0) {
					if (!result.empty()) result.append(" ");
					result.append((const char*)pdnsIt->Data.TXT.pStringArray[i]);
				}
			}
		}
		pdnsIt = pdnsIt->pNext;
	}

	DnsRecordListFree(pdns, DnsFreeRecordList);
	return true;
#else //! defined EMPTY_RESOLVER
	result.clear();
	return false;
#endif //! defined EMPTY_RESOLVER
}

#else
/*
 * initialize thread-safe m_res structure
 */
Resolver::Resolver()
{
#if ! defined EMPTY_RESOLVER
	memset((void*)&m_res, '\0', sizeof(m_res));
#ifdef HAS_RES_NINIT
	res_ninit(&m_res);
#else
	res_init();
	m_res = _res;
#endif
#endif //! defined EMPTY_RESOLVER
}

/*
 * close thread-safe m_res structure
 */
Resolver::~Resolver()
{
#if ! defined EMPTY_RESOLVER
#ifdef HAS_RES_NINIT
	res_nclose(&m_res);
#else
	res_close();
#endif
#endif //! defined EMPTY_RESOLVER
}

/*
 * request for the T_TXT record of an domain name, if an error occures (false is returned)
 *  else true is returned (regardless if the domain txt record exists or not)
 */
bool Resolver::GetTXT(const std::string& domain, std::string& result)
{
#if ! defined EMPTY_RESOLVER
	unsigned char answer[PACKETSZ];
	memset(answer, 0, PACKETSZ);

#ifdef HAS_RES_NINIT
	int answer_length = res_nquery(&m_res, domain.c_str(), C_IN, T_TXT, answer, PACKETSZ);
#else
	int answer_length = res_query(domain.c_str(), C_IN, T_TXT, answer, PACKETSZ);
#endif

	// Resolve failed
	if (answer_length < 0)
	{
#ifdef HAS_RES_NINIT
		int err = m_res.res_h_errno;
#else
		int err = h_errno;
#endif
		// permanent errors
		if (err == NO_DATA)
			return true;
		if (err == HOST_NOT_FOUND)
			return true;
		if (err == NO_RECOVERY)
			return true;

		// TRY_AGAIN
		return false;
	}
	// from here on, we will only return true
	// because we got whatever response..

	// Skip header
	unsigned char* answerptr = answer + sizeof(HEADER);

	// Skip request query...
	int qc = ntohs((unsigned short)((HEADER*)&answer)->qdcount);
	int x, s;
	for (x = 0, s = dn_skipname(answerptr, answer + answer_length);
			x < qc && s >= 0;
			++x, s = dn_skipname(answerptr, answer + answer_length))
	{
		if (s >= 0)
			answerptr += s + QFIXEDSZ;
	}

	if (qc == x)
	{
		int cc = ntohs((unsigned short)((HEADER*)&answer)->ancount);
		for (int i = 0; i < cc; ++i)
		{
			if ((s=dn_skipname(answerptr, answer + answer_length)) < 0)
			{
				return true;
			}
			int t;
			answerptr += s;
			GETSHORT(t, answerptr);
			answerptr += INT16SZ;
			answerptr += INT32SZ;
			GETSHORT(s, answerptr);
			switch(t)
			{
				case T_TXT:
					{
						const unsigned char* ptr = answerptr;
						size_t rec_len_left = s;

						while (rec_len_left > 0)
						{
							size_t txt_len = *ptr;
							if (txt_len > rec_len_left)
								break;

							if (!result.empty()) result.append(" ");
							result.append((const char*)ptr+1, txt_len);

							rec_len_left -= txt_len + 1;
							ptr += txt_len + 1;
						}
						answerptr += s;
					}
					break;
				default:
					// Skip eg. C_NAME
					answerptr += s;
			}
		}
		return true;
	}
	return true;
#else //! defined EMPTY_RESOLVER
	result.clear();
	return false;
#endif //! defined EMPTY_RESOLVER
}

#endif

}}
