/*
 *
 * Copyright (C) 2013 Kerio Technologies s.r.o
 *
 * This file is part of libdkim++.
 *
 * libdkim++ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libdkim++ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with libdkim++.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "Signatory1.hpp"
#include "Canonicalization.hpp"
#include "Base64.hpp"
#include "Util.hpp"

#include <algorithm>
#include <map>

namespace DKIM {

Signatory1::Signatory1(const ISignatoryOptions* options, bool doubleDots) declareThrow (DKIM::PermanentError)
: m_doubleDots(doubleDots), m_options(options), m_addedHeader(false)
{
	if (options == 0) {
		throw DKIM::TemporaryError("Null signatory options.");
	}
}

Signatory1::~Signatory1()
{
}

void Signatory1::beginSignature(std::istream& file) declareThrow (DKIM::PermanentError)
{
	while (m_msg.ParseLine(file, m_doubleDots) && !m_msg.IsDone()) { }

	EVP_MD_CTX ctx_body;
	EVP_MD_CTX_init( &ctx_body );

	// create signature for our body (message data)
	switch (m_options->GetAlgorithm())
	{
	case DKIM::DKIM_A_SHA1:
		EVP_DigestInit( &ctx_body, EVP_sha1() );
		break;
	case DKIM::DKIM_A_SHA256:
		EVP_DigestInit( &ctx_body, EVP_sha256() );
		break;
	}

	Conversion::CanonicalizationBody canonicalbody ( m_options->GetCanonModeBody() );

	// if we should limit the size of the body we hash
	bool limitBody = m_options->GetBodySignLength();
	unsigned long bodySize = m_options->GetBodyLength();

	// if we have a message: seek to GetBodyOffset()
	if (m_msg.GetBodyOffset() != -1)
	{
		file.clear();
		file.seekg(m_msg.GetBodyOffset());

		if ( ! file) {
			throw DKIM::PermanentError("Seek to body begin failed");
		}

		std::string s;
		while (std::getline(file, s) || file.peek() != EOF)
		{
			// double dots (postfix file may have .., instead of .)
			if (m_doubleDots && s.substr(0, 2) == "..")
			{
				s.erase(0, 1);
			}

			// remove possible \r (if not removed by getline *probably not*)
			if (s.size() > 0 && s[s.size()-1] == '\r')
				s.erase(s.size()-1);

			// canonical body
			std::vector<std::string> output;
			if (canonicalbody.FilterLine(s, output))
			{
				for (std::vector<std::string>::const_iterator i = output.begin();
					i != output.end(); ++i)
				{
					if (limitBody && bodySize == 0) break;
					EVP_DigestUpdate(&ctx_body, i->c_str(),
						limitBody?_MIN(i->size(), bodySize):i->size());
					bodySize -= (unsigned long)_MIN(i->size(), bodySize);
				}
			}
		}
	}

	// else call (Done) -- which may insert a last CRLF if the body was empty
	if (m_msg.GetBodyOffset() == -1 || file.peek() == EOF)
	{
		std::vector<std::string> output;
		if (canonicalbody.Done(output))
		{
			for (std::vector<std::string>::const_iterator i = output.begin();
				i != output.end(); ++i)
			{
				if (limitBody && bodySize == 0) break;
				EVP_DigestUpdate(&ctx_body, i->c_str(),
					limitBody?_MIN(i->size(), bodySize):i->size());
				bodySize -= (unsigned long)_MIN(i->size(), bodySize);
			}
		}
	}
	if (limitBody && bodySize > 0)
		throw DKIM::PermanentError("Body sign limit exceed the size of the canonicalized message length");

	unsigned char md_value[EVP_MAX_MD_SIZE];
	unsigned int md_len;
	EVP_DigestFinal_ex(&ctx_body, md_value, &md_len);
	EVP_MD_CTX_cleanup( &ctx_body );

	std::string tmp;
	tmp.assign((char*)md_value, md_len);
	m_bodySignature = Conversion::Base64().Encode(tmp);

	m_dkimSignature = computeDkimSignature();
}

void Signatory1::addHeaderToFront(const std::string& header)
{
	m_msg.AddHeaderToFront(header, m_doubleDots);
	m_addedHeader = true;
}

std::string Signatory1::computeDkimSignature() declareThrow (DKIM::PermanentError)
{
	EVP_MD_CTX ctx_head;
	EVP_MD_CTX_init( &ctx_head );

	// create signature for our header
	switch (m_options->GetAlgorithm())
	{
	case DKIM::DKIM_A_SHA1:
		EVP_SignInit( &ctx_head, EVP_sha1() );
		break;
	case DKIM::DKIM_A_SHA256:
		EVP_SignInit( &ctx_head, EVP_sha256() );
		break;
	}

	Conversion::CanonicalizationHeader canonicalhead ( m_options->GetCanonModeHeader() );

	std::list<std::string> headersToSign = m_options->GetHeaders();
	std::list<std::string> signedHeaders;

	bool signAll = false;
	if (headersToSign.empty()) signAll = true;

	// add all headers to our cache (they will be pop of the end)
	std::map<std::string, Message::HeaderList> headerCache;
	for (Message::HeaderList::const_iterator hIter = m_msg.GetHeaders().begin();
		hIter != m_msg.GetHeaders().end();
		++hIter)
	{
		std::string name = (*hIter)->GetName();
		transform(name.begin(), name.end(), name.begin(), tolower);

		headerCache[name].push_back(*hIter);
		if (signAll)
			headersToSign.push_back(name);
	}

	while (!headersToSign.empty())
	{
		std::string tmp;
		std::string name = headersToSign.front();
		if (!name.empty())
		{
			transform(name.begin(), name.end(), name.begin(), tolower);

			std::map<std::string, Message::HeaderList>::iterator head = headerCache.find(name);
			if (head == headerCache.end() || head->second.empty())
			{
				if (m_options->GetSignEmptyHeaders())
					tmp = canonicalhead.FilterHeader(name + ":\r\n");
			} else {
				tmp = canonicalhead.FilterHeader(head->second.back()->GetHeader()) + "\r\n";
				head->second.pop_back();
			}
			if (!tmp.empty())
			{
				EVP_SignUpdate(&ctx_head, tmp.c_str(), tmp.size());
				signedHeaders.push_back(name);
			}
		}
		headersToSign.pop_front();
	}

	std::string dkimHeader;
	dkimHeader += "DKIM-Signature: v=1; a=" + Util::Algorithm2String(m_options->GetAlgorithm()) + "; c="
		+ Util::CanonMode2String(m_options->GetCanonModeHeader()) + "/" + Util::CanonMode2String(m_options->GetCanonModeBody()) + ";\r\n";

	char* limit = 0x0;
	if (m_options->GetBodySignLength()) {
#if defined _MSC_VER
		limit = (char*)malloc(32);	// 32 is enough for 32 bit number as string
		sprintf(limit, "%lu", m_options->GetBodyLength());
#else
		asprintf(&limit, "%lu", m_options->GetBodyLength());
#endif
	}

	dkimHeader += "\td=" + m_options->GetDomain() + "; s=" + m_options->GetSelector() +
		(!limit?"":"; l=" + std::string(limit))
		+ ";\r\n";
	free(limit);

	std::string headerlist = "\th=";
	for (std::list<std::string>::const_iterator i = signedHeaders.begin();
		i != signedHeaders.end(); ++i)
	{
		bool insertColon = (i == signedHeaders.begin())?false:true;
		if (headerlist.size() + i->size() + (insertColon?1:0) > 80)
		{
			dkimHeader += headerlist + (insertColon?":":"") + "\r\n";
			headerlist = "\t " + *i;
		} else {
			headerlist += (insertColon?":":"") + *i;
		}
	}
	dkimHeader += headerlist + ";\r\n";
	dkimHeader += "\tbh=" + m_bodySignature + ";\r\n";
	dkimHeader += "\tb=";

	std::string tmp2 = canonicalhead.FilterHeader(dkimHeader);
	EVP_SignUpdate(&ctx_head, tmp2.c_str(), tmp2.size());

	unsigned int len;
	const size_t dufLen = EVP_PKEY_size(m_options->GetPrivateKey());
	unsigned char* data = (unsigned char*)malloc(dufLen);
	if (EVP_SignFinal(&ctx_head,
		data,
		&len,
		m_options->GetPrivateKey()
		) != 1)
	{
		free(data);
		throw DKIM::PermanentError("Message could not be signed");
	}
	EVP_MD_CTX_cleanup( &ctx_head );

	std::string tmp3; tmp3.assign((const char*)data, len);
	free(data);

	int offset = 3; // "\tb=";
	std::string split = Conversion::Base64().Encode(tmp3);
	while (!split.empty())
	{
		dkimHeader += split.substr(0, 80 - offset);
		split.erase(0, 80 - offset);
		if (!split.empty())
			dkimHeader += "\r\n\t ";
		offset = 2; // "\t ";
	}

	return dkimHeader;
}

std::string Signatory1::getSignatureHeader() declareThrow (DKIM::PermanentError)
{
	if (m_addedHeader) {
		return computeDkimSignature();
	}
	else {
		return m_dkimSignature;
	}
}

Signatory1::Signatory1(const Signatory1& s)
:	m_msg(s.m_msg),
	m_doubleDots(s.m_doubleDots),
	m_options(s.m_options),
	m_bodySignature(s.m_bodySignature),
	m_dkimSignature(s.m_dkimSignature),
	m_addedHeader(s.m_addedHeader)
{
}

ISignatory* Signatory1::clone() const
{
	return new Signatory1(*this);
}

}
